/* nppt.c - format input as paper tape in a variety of ways.  */
/* RCS: $Id: nppt.c,v 1.17 2004/06/06 12:18:02 jsm28 Exp $ */

/* Copyright (c) 1998, 1999, 2000, 2004 Joseph Samuel Myers.
 * All rights reserved.
 *
 * This file is part of nppt, a program to format its input in a
 * representation of paper tape.
 *
 * Nppt is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * Nppt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with nppt; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <errno.h>

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "nppt.h"
#include "util.h"

const struct tape_operations *tape;
const void *tape_params;
unsigned int tape_max;
unsigned int tape_bits;

static const struct tape_type *const tape_types[] = {
  vert_tape_types,
  horiz_tape_types,
  ps_tape_types,
  NULL
};

int main(int argc, char **argv); /* Main program */

static void start_tape(void);
static void print_cmd_arg(const char *arg);
static void print_cmd_arg_space(void);
static void print_stdin_char(int c);
static void end_tape(void);

static void translate_and_print_char(int c);
static void translation_end(void);
static void print_tape_char(unsigned int c, int character);



int show_text_flag = 0; /* Show text alongside the tape?  */
unsigned long screen_columns; /* Number of screen columns (if needed) */
int override_screen_columns_flag = 0; /* Override the number of screen
					 columns from the command line? */

enum parity_types {
  no_parity,
  zero_bit_parity,
  one_bit_parity,
  odd_parity,
  even_parity
};

enum parity_types parity = no_parity;
int parity_place = -1;

/* main - main program */
int
main(int argc, char **argv)
{
  const char *tape_type = NULL;
  /* Get the program name for error and usage messages */
  get_program_name(argv[0]);
  /* Parse options */
  while (1) {
#ifdef HAVE_GETOPT_LONG
    const struct option long_options[] = {
      { "help", no_argument, (int *)NULL, 'h' },
      { "version", no_argument, (int *)NULL, 'v' },
      { "tape-type", required_argument, (int *)NULL, 't' },
      { "show-text", no_argument, (int *)NULL, 'T' },
      { "columns", required_argument, (int *)NULL, 'c' },
      { "parity", required_argument, (int *)NULL, 'p' },
      { "parity-bit", required_argument, (int *)NULL, 'P' },
      { NULL, 0, NULL, 0 }
    };
    int option_index;
#endif
    const char short_options[] = "hvt:Tc:p:P:";
    int ret;
#ifdef HAVE_GETOPT_LONG
    ret = getopt_long(argc, argv, short_options, long_options, &option_index);
#else
    ret = getopt(argc, argv, short_options);
#endif
    if (ret == -1)
      break; /* End of options */
    switch (ret) {
    case 'h':
      /* Help */
      xprintf("Usage: %s [ --help | --version ] [ --tape-type=TYPE]\n"
	      "\t[ --show-text ] [ --columns=NUMBER ] [ --parity=[01oe] ]\n"
	      "\t[ --parity-bit=NUMBER ] [ WORDS ]\n",
	      program_name);
      xputs("\nRecognised tape types:\n\n");
      {
	const struct tape_type *const *i, *j;
	for (i = tape_types; *i != NULL; i++)
	  for (j = *i; j->name != NULL; j++)
	    xprintf("%s\t%s%s\n", j->name, j->desc,
		    (strcmp(j->name, DEFAULT_FORMAT) ? "" : " (default)"));
      }
      xputs("\nReport bugs to jsm@polyomino.org.uk.\n");
      xflush(stdout, "standard output");
      exit(EXIT_SUCCESS);
    case 'v':
      /* Version */
      xputs(VERSION);
      xputs("Copyright (c) 2004 Joseph Samuel Myers.\n"
	    "Nppt comes with ABSOLUTELY NO WARRANTY.\n"
	    "You may redistribute copies of nppt under the terms of the\n"
	    "GNU General Public License.  For more information, see the\n"
	    "file COPYING in the source distribution.\n");
      xflush(stdout, "standard output");
      exit(EXIT_SUCCESS);
    case 't':
      if (tape_type == NULL)
	tape_type = optarg;
      else
	die("multiple --tape-type or -t options");
      break;
    case 'T':
      if (show_text_flag)
	die("multiple --show-text or -T options");
      else
	show_text_flag = 1;
      break;
    case 'c':
      if (override_screen_columns_flag) {
	die("multiple --columns or -c options");
      } else {
	char *endptr;
	override_screen_columns_flag = 1;
	if (strtol(optarg, NULL, 0) < 0)
	  die("a negative number of columns does not make sense");
	errno = 0;
	screen_columns = strtoul(optarg, &endptr, 0);
	if (errno != 0)
	  die("--columns or -c value '%s': %s", optarg, strerror(errno));
	if ((endptr == optarg) || (*endptr != '\0'))
	  die("--columns or -c value '%s' is not numeric or has trailing junk",
	      optarg);
      }
      break;
    case 'p':
      if (parity != no_parity) {
	die("multiple --parity or -p options");
      } else {
	if (strlen(optarg) != 1) {
	  die("--parity or -p option argument '%s' is invalid", optarg);
	}
	switch (optarg[0]) {
	case '0':
	  parity = zero_bit_parity;
	  break;
	case '1':
	  parity = one_bit_parity;
	  break;
	case 'o':
	  parity = odd_parity;
	  break;
	case 'e':
	  parity = even_parity;
	  break;
	default:
	  die("--parity or -p option argument '%s' is invalid", optarg);
	}
      }
      break;
    case 'P':
      if (parity_place != -1) {
	die("multiple --parity-bit or -P options");
      } else {
	long ppl;
	char *endptr;
	errno = 0;
	ppl = strtol(optarg, &endptr, 0);
	if (errno != 0)
	  die("--parity-bit or -P value '%s': %s", optarg, strerror(errno));
	if ((endptr == optarg) || (*endptr != '\0'))
	  die("--parity-bit or -P value '%s' is not numeric"
	      " or has trailing junk", optarg);
	if (ppl < 0)
	  die("a negative bit position does not make sense");
	parity_place = ppl;
	if (parity_place != ppl)
	  die("parity bit position too big");
      }
      break;
    case '?':
      /* Bad option - error message printed */
      exit(EXIT_FAILURE);
    default:
      die("panic: bad return %d from getopt_long", ret);
    }
  }
  /* Now determine the type of tape display wanted.  */
  if (tape_type == NULL)
    tape_type = DEFAULT_FORMAT;
  {
    const struct tape_type *const *i, *j;
    for (i = tape_types; *i != NULL; i++) {
      for (j = *i; j->name != NULL; j++) {
	if (!strcmp(j->name, tape_type)) {
	  tape = &j->ops;
	  tape_params = j->params;
	  tape_max = j->max;
	  tape_bits = j->bits;
	  goto found_name;
	  break;
	}
      }
    }
    die("tape type '%s' not known", tape_type);
  }
 found_name: ;
  /* Check the parity bit position makes sense */
  if ((parity_place != -1) && (parity == no_parity))
    warning("parity bit position specification ignored since no parity");
  if (parity_place == -1)
    parity_place = tape_bits - 1;
  if ((unsigned int)parity_place >= tape_bits)
    die("parity bit position too big");
  /* If we have any non-option arguments then we print those as paper tape;
   * otherwise we read standard input.  */
  if (optind < argc) {
    int i;
    start_tape();
    for (i = optind; i < argc; i++) {
      print_cmd_arg(argv[i]);
      if (i < argc - 1)
	print_cmd_arg_space();
    }
    end_tape();
  } else {
    start_tape();
    while (1) {
      int c;
#ifdef HAVE_GETC_UNLOCKED
      c = getc_unlocked(stdin);
#else
      c = getc(stdin);
#endif
      if (c == EOF) {
	if (ferror(stdin))
	  die("error reading stdin: %s", strerror(errno));
	else {
	  /* EOF */
	  end_tape();
	  break;
	}
      }
      print_stdin_char(c);
    }
  }
  exit(EXIT_SUCCESS);
}

/* start_tape - start printing the tape */
static void
start_tape(void)
{
  tape->start_tape();
}

/* print_cmd_arg - print something on the tape from a command line argument.
 * Maybe do character set translation.  */
static void
print_cmd_arg(const char *arg)
{
  const char *c;
  for (c = arg; *c; c++)
    translate_and_print_char((unsigned char)*c);
}

/* print_cmd_arg_space - print a space corresponding to the gap between
 * command line arguments.  Maybe do character set translation.  */
static void
print_cmd_arg_space(void)
{
  translate_and_print_char((unsigned char)' ');
}

/* print_stdin_char - print a character from stdin.  Maybe do character set
 * translation.  */
static void
print_stdin_char(int c)
{
  translate_and_print_char(c);
}

/* end_tape - end printing the tape */
static void
end_tape(void)
{
  translation_end();
  tape->end_tape();
}

/* translate_and_print_char - translate a character on input (possibly
 * to more than one character), and print it.  */
static void
translate_and_print_char(int c)
{
  unsigned int tc;
  /* No translation implemented yet, but we do implement parity.  */
  tc = (unsigned int)c;
  if (parity != no_parity) {
    /* Apply the appropriate parity to tc */
    if (tc & (1 << parity_place))
      die("can't print character %u, parity bit already used", tc);
    switch (parity) {
    case no_parity:
      break; /* Can't actually happen */
    case zero_bit_parity:
      break;
    case one_bit_parity:
      tc |= (1 << parity_place);
      break;
    case odd_parity:
    case even_parity:
      /* Determine the parity we have, and that we want */
      {
	unsigned int wanted_parity = (parity == odd_parity ? 1 : 0);
	unsigned int pbit;
	unsigned int tp;
	/* Is this the best way of getting parity (we assume <= 32bit tape)? */
	tp = tc;
	tp = (tp & 0x0000ffff) ^ ((tp & 0xffff0000) >> 16);
	tp = (tp & 0x00ff) ^ ((tp & 0xff00) >> 8);
	tp = (tp & 0x0f) ^ ((tp & 0xf0) >> 4);
	tp = (tp & 0x3) ^ ((tp & 0xc) >> 2);
	tp = (tp & 0x1) ^ ((tp & 0x2) >> 1);
	pbit = tp ^ wanted_parity;
	tc = tc | (pbit << parity_place);
      }
    }
  }
  print_tape_char(tc, c);
}

/* translation_end - print any final characters (end of shift state, etc.)
 * required by the character set translation.  */
static void
translation_end(void)
{
  /* Nothing do be done here currently */
}

/* print_tape_char - print a character on the tape.  Here c is the raw
 * character for the tape; translations may have been done previously.
 * 'character' is the character c represents.  */
static void
print_tape_char(unsigned int c, int character)
{
  if (c > tape_max)
    die("can't print character %u, too wide for tape", c);
  tape->print_character(c, character);
}
