/* tape-horiz.c - format input as paper tape, horizontal ASCII art.  */
/* RCS: $Id: tape-horiz.c,v 1.3 1999/01/11 00:24:34 jsm28 Exp $ */

/* Copyright (c) 1998, 1999 Joseph Samuel Myers.  All rights reserved.
 *
 * This file is part of nppt, a program to format its input in a
 * representation of paper tape.
 *
 * Nppt is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * Nppt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with nppt; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <ctype.h>
#include <stdio.h>

#include "nppt.h"
#include "util.h"
#include "winsize.h"

struct horiz_tape_params {
  size_t height; /* Height of tape */
  const char *const *left; /* Strings for each left column */
  const char *noholes; /* A column (no holes) */
  const char *holes; /* A column (holes) */
  const char *const *right; /* Right part */
  const size_t *hole_bits; /* Bit for each place */
  size_t text_place; /* Characters from bottom text can be displayed
		      * (maybe 0) */
};

static const size_t horiz8_tape_bits[] = { 7, 6, 5, 4, 3, 2, 1, 0 };
static const char *const horiz8_left[] = {
  "      <    ",
  "     /.\\   ",
  "    / . \\  ",
  "   /  .  \\ ",
  "  /   .   -",
  " /    .   -",
  "-     .   -",
  NULL
};
static const char *const horiz8_right[] = {
  "-     .   -",
  "-     .   -",
  "-     <   -",
  "-    / \\  -",
  "-   /   \\ -",
  "-  /     \\-",
  "- /        ",
  "-/         ",
  NULL
};
static const struct horiz_tape_params horiz8_tape_params = {
  11,
  horiz8_left,
  "-     .   -",
  "-OOOOO.OOO-",
  horiz8_right,
  horiz8_tape_bits,
  0
};

static const size_t horiz7_tape_bits[] = { 6, 5, 4, 3, 2, 1, 0 };
static const char *const horiz7_left[] = {
  "     <    ",
  "    /.\\   ",
  "   / . \\  ",
  "  /  .  \\ ",
  " /   .   -",
  "-    .   -",
  NULL
};
static const char *const horiz7_right[] = {
  "-    .   -",
  "-    .   -",
  "-    <   -",
  "-   / \\  -",
  "-  /   \\ -",
  "- /     \\-",
  "-/        ",
  NULL
};
static const struct horiz_tape_params horiz7_tape_params = {
  10,
  horiz7_left,
  "-    .   -",
  "-OOOO.OOO-",
  horiz7_right,
  horiz7_tape_bits,
  0
};

static const size_t horiz6_tape_bits[] = { 5, 4, 3, 2, 1, 0 };
static const char *const horiz6_left[] = {
  "    <    ",
  "   /.\\   ",
  "  / . \\  ",
  " /  .  \\ ",
  "-   .   -",
  NULL
};
static const char *const horiz6_right[] = {
  "-   .   -",
  "-   .   -",
  "-   <   -",
  "-  / \\  -",
  "- /   \\ -",
  "-/     \\-",
  NULL
};
static const struct horiz_tape_params horiz6_tape_params = {
  9,
  horiz6_left,
  "-   .   -",
  "-OOO.OOO-",
  horiz6_right,
  horiz6_tape_bits,
  0
};

static const size_t horiz5_tape_bits[] = { 4, 3, 2, 1, 0 };
static const char *const horiz5_left[] = {
  "   <    ",
  "  /.\\   ",
  " / . \\  ",
  "-  .  \\ ",
  "-  .   -",
  NULL
};
static const char *const horiz5_right[] = {
  "-  .   -",
  "-  .   -",
  "-  <   -",
  "- / \\  -",
  "-/   \\ -",
  "      \\-",
  NULL
};
static const struct horiz_tape_params horiz5_tape_params = {
  8,
  horiz5_left,
  "-  .   -",
  "-OO.OOO-",
  horiz5_right,
  horiz5_tape_bits,
  0
};

static void horiz_start_tape(void);
static void horiz_print_char(unsigned int, int);
static void horiz_end_tape(void);
static void horiz_add_str(const char *);

const struct tape_type horiz_tape_types[] = {
  { "horiz8", "Horizontal ASCII art 8-track tape", 255U, 8U,
    { horiz_start_tape, horiz_print_char, horiz_end_tape },
    &horiz8_tape_params
  },
  { "horiz7", "Horizontal ASCII art 7-track tape", 127U, 7U,
    { horiz_start_tape, horiz_print_char, horiz_end_tape },
    &horiz7_tape_params
  },
  { "horiz6", "Horizontal ASCII art 6-track tape", 63U, 6U,
    { horiz_start_tape, horiz_print_char, horiz_end_tape },
    &horiz6_tape_params
  },
  { "horiz5", "Horizontal ASCII art 5-track tape", 31U, 5U,
    { horiz_start_tape, horiz_print_char, horiz_end_tape },
    &horiz5_tape_params
  },
  { NULL, NULL, 0U, 0U,
    { NULL, NULL, NULL },
    NULL
  }
};

static const struct horiz_tape_params *htp;
static char *horiz_current;
static char **horiz_buffers;
static size_t horiz_buffers_allocated;
static size_t horiz_buffers_pos;

/* horiz_start_tape - start printing the tape, horizontal ASCII art */
static void
horiz_start_tape(void)
{
  const char *const *i;
  htp = (const struct horiz_tape_params *)tape_params;
  if (!override_screen_columns_flag) {
    struct window_size win;
    get_window_size(&win);
    screen_columns = win.cols;
  }
  /* Allocate buffers */
  {
    size_t amt;
    size_t j;
    amt = (screen_columns ? (screen_columns + 2) : 1024);
    horiz_buffers = xmalloc((htp->height + 1) * sizeof(char *));
    horiz_current = xmalloc(htp->height + 2);
    for (j = 0; j < htp->height + 1; j++)
      horiz_buffers[j] = xmalloc(amt);
    horiz_buffers_allocated = amt;
    horiz_buffers_pos = 0;
  }
  for (i = htp->left; *i != NULL; i++)
    horiz_add_str(*i);
}

/* horiz_print_char - print a character, horizontal ASCII art */
static void
horiz_print_char(unsigned int c, int character)
{
  size_t p, q;
  char *hcp;
  q = 0;
  hcp = horiz_current;
  for (p = 0; p <= htp->height; p++) {
    char c1, c2;
    if (show_text_flag && p == (htp->height - htp->text_place)) {
      int pc2;
      pc2 = (isprint(character) ? character : (unsigned char)' ');
      *(hcp++) = pc2;
    }
    if (p == htp->height) {
      *(hcp++) = '\0';
      break;
    }
    c1 = htp->noholes[p];
    c2 = htp->holes[p];
    if (c1 == c2) {
      *(hcp++) = c1;
    } else {
      *(hcp++) = ((c & (1 << htp->hole_bits[q++])) ? c2 : c1);
    }
  }
  horiz_add_str(horiz_current);
}

/* horiz_end_tape - end printing the tape, horizontal ASCII art */
static void
horiz_end_tape(void)
{
  const char *const *i;
  for (i = htp->right; *i != NULL; i++)
    horiz_add_str(*i);
  if (horiz_buffers_pos > 0) {
    size_t j;
    for (j = 0; j < htp->height + 1; j++) {
      horiz_buffers[j][horiz_buffers_pos] = '\n';
      horiz_buffers[j][horiz_buffers_pos + 1] = '\0';
      xputs(horiz_buffers[j]);
    }
  }
  xflush(stdout, "standard output");
  /* Things will get freed when the program exits */
}

/* horiz_add_str - add a string to a horizontal tape */
static void
horiz_add_str(const char *str)
{
  size_t i, j;
  for (i = 0; str[i] != '\0'; i++) {
    horiz_buffers[i][horiz_buffers_pos] = str[i];
  }
  for (j = i; j < htp->height + 1; j++)
    horiz_buffers[j][horiz_buffers_pos] = ' ';
  horiz_buffers_pos++;
  if (horiz_buffers_pos == horiz_buffers_allocated - 2) {
    if (screen_columns) {
      /* Print this part of the tape */
      size_t k;
      for (k = 0; k < htp->height + 1; k++) {
	horiz_buffers[k][horiz_buffers_pos] = '\n';
	horiz_buffers[k][horiz_buffers_pos + 1] = '\0';
	xputs(horiz_buffers[k]);
      }
      if (show_text_flag)
	xputc((unsigned char)'\n', stdout, "standard output");
      horiz_buffers_pos = 0;
    } else {
      /* Reallocate */
      size_t k;
      horiz_buffers_allocated *= 2;
      for (k = 0; k < htp->height + 1; k++)
	horiz_buffers[k] = xrealloc(horiz_buffers[k], horiz_buffers_allocated);
    }
  }
}
