/* tape-ps.c - format input as paper tape, Encapsulated PostScript.  */
/* RCS: $Id: tape-ps.c,v 1.6 2004/06/06 12:46:52 jsm28 Exp $ */

/* Copyright (c) 1996, 1997, 1998, 1999, 2004 Joseph Samuel Myers.
 * All rights reserved.
 *
 * This file is part of nppt, a program to format its input in a
 * representation of paper tape.
 *
 * Nppt is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * Nppt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with nppt; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * As a special exception, if the PostScript code contained in this file
 * is included in an output file from this program, this does not by itself
 * cause the output file to be covered by the GNU General Public License.
 * This exception does not however invalidate any other reasons why the
 * output file might be covered by the GNU General Public License.
 *
 */

#include <math.h>
#include <stdio.h>
#include <time.h>

#include "nppt.h"
#include "util.h"

#define MM_TO_PT(len)	((len) * (72.0 / 25.4))
#define PT_TO_MM(len)	((len) * (25.4 / 72.0))
#define DEG_TO_RAD(angle)	((angle) * (3.14159265358979323846 / 180.0))

struct ps_tape_params {
  double tape_width; /* Tape width, in mm */
  double sprocket_height; /* Sprocket holes (centre), offset from
			   * reference (bottom) edge, in mm */
  int holes_below_sprocket; /* Number of holes below the sprocket hole */
  int character_bits; /* Total number of bits in a character */
  double track_spacing; /* Spacing of tracks (in mm) */
  double sprocket_diameter; /* Diameter of sprocket hole (in mm) */
  double data_diameter; /* Diameter of data hole (in mm) */
  double data_hole_offset; /* Offset of centre of data holes to the right of
			    * centre of sprocket holes (used on typesetter
			    * (TS) tape) (in mm) */
  double inter_character_space; /* Inter-character spacing, in mm */
  double upper_tearoff_angle; /* Angle of upper tearoff in radians
			       * (counterclockwise from the rightward
			       * horizontal) */
  double lower_tearoff_angle; /* Angle of lower tearoff in radians
			       * (clockwise from the rightward horizontal) */
  double line_thickness; /* Thickness of the lines used to draw the tape,
			  * in mm */
};

/* Parameters from Facit 4070 manual (full reference?????).
 * ECMA-10 has the same parameters for the 25.40 and 17.46 mm tapes
 * (the only ones it covers), but some slightly different tolerances.
 */

static const struct ps_tape_params ps8_tape_params = {
  25.40 /* +/- 0.08 */,
  9.96 /* +/- 0.1 */,
  3,
  8,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.0,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

static const struct ps_tape_params ps7_tape_params = {
  22.23 /* +/- 0.08 */,
  9.96 /* +/- 0.1 */,
  3,
  7,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.0,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

static const struct ps_tape_params ps6_tape_params = {
  22.23 /* +/- 0.08; this _is_ the same as for 7-track */,
  9.96 /* +/- 0.1 */,
  3,
  6,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.0,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

static const struct ps_tape_params ps5_tape_params = {
  17.46 /* +/- 0.08 */,
  9.96 /* +/- 0.1 */,
  3,
  5,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.0,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

static const struct ps_tape_params psTS_tape_params = {
  22.23 /* +/- 0.08 */,
  11.02 /* +/- 0.1 */,
  3,
  6,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.33 /* +/- 0.02; the effect of this offset is to align the _leading edge_
	* of the sprocket and data holes, not the centres */,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

static void ps_start_tape(void);
static void ps_print_char(unsigned int, int);
static void ps_end_tape(void);

const struct tape_type ps_tape_types[] = {
  { "PS8", "Standard 8-track tape (Encapsulated PostScript output)", 255U, 8U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &ps8_tape_params
  },
  { "PS7", "Standard 7-track tape (Encapsulated PostScript output)", 127U, 7U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &ps7_tape_params
  },
  { "PS6", "Standard 6-track tape (Encapsulated PostScript output)", 63U, 6U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &ps6_tape_params
  },
  { "PS5", "Standard 5-track tape (Encapsulated PostScript output)", 31U, 5U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &ps5_tape_params
  },
  { "PSTS", "6-track typesetter (TS) tape (Encapsulated PostScript output)",
    63U, 6U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &psTS_tape_params
  },
  { NULL, NULL, 0U, 0U,
    { NULL, NULL, NULL },
    NULL
  }
};

static const struct ps_tape_params *pstp;
static size_t ps_number_of_chars;
static size_t ps_chars_allocated;
static unsigned int *ps_bits;
static unsigned char *ps_chars;

/* ps_start_tape - start printing the tape, PostScript */
static void
ps_start_tape(void)
{
  /* In order to know the bounding box we need to have all the tape data;
   * and it isn't clear whether EPSF files can use (atend) for the bounding
   * box; so we defer all the work to the end.  */
  if (override_screen_columns_flag)
    warning("screen columns specification ignored for this tape design");
  pstp = (const struct ps_tape_params *)tape_params;
  ps_chars_allocated = 1024;
  ps_bits = xmalloc(ps_chars_allocated * sizeof(unsigned int));
  if (show_text_flag)
    ps_chars = xmalloc(ps_chars_allocated);
  ps_number_of_chars = 0;
}

/* ps_print_char - print a character, PostScript */
static void
ps_print_char(unsigned int c, int character)
{
  /* Actually with PostScript output we just store the character (both
   * the bit pattern and the original character, if appropriate), to output
   * at the end.  */
  if (ps_number_of_chars == ps_chars_allocated) {
    ps_chars_allocated *= 2;
    ps_bits = xrealloc(ps_bits, ps_chars_allocated * sizeof(unsigned int));
    if (show_text_flag)
      ps_chars = xrealloc(ps_chars, ps_chars_allocated);
  }
  if (show_text_flag)
    ps_chars[ps_number_of_chars] = character;
  ps_bits[ps_number_of_chars++] = c;
}

/* ps_end_tape - end printing the tape, PostScript */
static void
ps_end_tape(void)
{
  /* This is where we do all the work */
#define MM_TO_UNITS(len)	((len) * (1000.0 / pstp->tape_width))
#define MIN(x, y)	(((x) < (y)) ? (x) : (y))
#define MAX(x, y)	(((x) > (y)) ? (x) : (y))
  double line_thickness; /* Line thickness, in the 1000 unit scale */
  double font_base; /* Base of the font, in the 1000 unit scale */
  double font_top; /* Top of the font, in the 1000 unit scale */
  double norm_left, norm_right, norm_width; /* dimensions of normal
					     * character */
  double lend_left, lend_right, lend_width; /* left end */
  int lend_char_positions, lend_left_pos;
  double rend_left, rend_right, rend_width; /* right end */
  int rend_char_positions, rend_left_pos;
  double sprocket_radius, data_radius; /* radii of holes */
  double data_hole_offset;
  double sprocket_height;
  double track_spacing;
  double utheta, dtheta;
  line_thickness = MM_TO_UNITS(pstp->line_thickness);
  font_base = 0.0 - (line_thickness / 2.0);
  font_top = 1000.0 + (line_thickness / 2.0);
  norm_width = MM_TO_UNITS(pstp->inter_character_space);
  sprocket_radius = MM_TO_UNITS(pstp->sprocket_diameter) / 2.0;
  data_radius = MM_TO_UNITS(pstp->data_diameter) / 2.0;
  data_hole_offset = MM_TO_UNITS(pstp->data_hole_offset);
  sprocket_height = MM_TO_UNITS(pstp->sprocket_height);
  track_spacing = MM_TO_UNITS(pstp->track_spacing);
  norm_left = MIN(0, MIN(norm_width / 2.0 - sprocket_radius,
			 norm_width / 2.0 + data_hole_offset - data_radius));
  norm_right = MAX(norm_width,
		   MAX(norm_width / 2.0 + sprocket_radius,
		       norm_width / 2.0 + data_hole_offset + data_radius));
  utheta = pstp->upper_tearoff_angle;
  dtheta = pstp->lower_tearoff_angle;
  /* The left end.  We work out the space needed from hole centre at the
   * left to the inner intersection of the lines at top and bottom.  If
   * holes from a character extend to the left of its strip, we adjust the
   * right boundary accordingly.  If they extend to the right, we adjust the
   * left boundary accordingly.  */
  {
    double tmpl, tmpr;
    double tmpu, tmpd;
    tmpu = (((1000.0 - sprocket_height) / tan(utheta))
	    + ((line_thickness / 2.0) * (1 / sin(utheta) - 1 / tan(utheta))));
    tmpd = ((sprocket_height / tan(dtheta))
	    + ((line_thickness / 2.0) * (1 / sin(dtheta) - 1 / tan(dtheta))));
    tmpr = MAX(tmpu, tmpd) - norm_left;
    tmpl = -(line_thickness / 2.0) * ((cos(utheta) + cos(dtheta))
				      / sin(utheta + dtheta));
    tmpl -= (norm_right - norm_width);
    lend_left_pos = -floor(tmpl / norm_width + 0.5);
    lend_char_positions = ceil(tmpr / norm_width + 0.5) + lend_left_pos;
    lend_left = 0;
    lend_right = lend_width = norm_width * lend_char_positions;
  }
  /* The right end - adjustments as commented above */
  {
    double tmpl, tmpr;
    double tmpu, tmpd;
    tmpu = (((1000.0 - sprocket_height) / tan(utheta))
	    + ((line_thickness / 2.0) * (1 / tan(utheta) + 1 / sin(utheta))));
    tmpd = ((sprocket_height / tan(dtheta))
	    + ((line_thickness / 2.0) * (1 / tan(dtheta) + 1 / sin(dtheta))));
    tmpr = MAX(tmpu, tmpd) - norm_left;
    tmpl = -(line_thickness / 2.0) * ((cos(utheta) + cos(dtheta))
				      / sin(utheta + dtheta));
    tmpl -= (norm_right - norm_width);
    rend_left_pos = -floor(tmpl/norm_width + 0.5);
    rend_char_positions = ceil(tmpr / norm_width + 0.5) + rend_left_pos;
    rend_left = 0;
    rend_right = rend_width = norm_width * rend_char_positions;
  }
  /* The EPSF header */
  {
    int llx, lly, urx, ury;
    time_t t;
    llx = 0;
    if (show_text_flag) {
      /* Courier characters can extend to 216/1000 below their box */
      lly = floor(-MM_TO_PT(pstp->inter_character_space) / 0.6 * 1.216);
    } else {
      lly = floor(-MM_TO_PT(pstp->line_thickness) / 2.0);
    }
    ury = ceil(72.0 + MM_TO_PT(pstp->line_thickness) / 2.0);
    urx = ceil((ps_number_of_chars + lend_char_positions + rend_char_positions)
	       * MM_TO_PT(pstp->inter_character_space));
    t = time(NULL);
    xprintf("%%!PS-Adobe-3.0 EPSF-3.0\n"
	    "%%%%BoundingBox: %d %d %d %d\n"
	    "%%%%LanguageLevel: 2\n"
	    "%%%%Creator: "
	    VERSION
	    "%%%%CreationDate: %s",
	    llx, lly, urx, ury, ctime(&t));
    if (show_text_flag)
      xputs("%%DocumentNeededResources: font Courier\n");
    xputs("\n\n");
  }
  /* The font */
  {
    int fontllx, fontlly, fonturx, fontury;
    int normllx, normurx;
    int r_urx, l_urx;
    /* This font was originally designed in 1996 for the Cambridge
     * University Computer Preservation Society, whose logo consists
     * of a representation of the Society's name on paper tape.  Some
     * of the C comments below are derived from those in the original
     * PostScript version.
     *
     * The font defines the PaperTape8 (or PaperTape7, etc.) font, and the
     * procedures scaletape and (if appropriate) scaletext to scale the
     * fonts and set the current font.
     *
     * The method of using a font and printing strings is fundamentally
     * limited to 8-track tapes, but I don't know of wider versions.
     *
     * In 12pt Courier, characters are exactly(?) 0.1in across; this is
     * where the 0.6 scale factor used in various places comes from.  */
    xputs("8 dict begin\n");
    /* The EPSF spec notes that we should initialise all variables before
     * binding a procedure that uses them, to avoid problems if the
     * including document uses the same names for operators.  Since this
     * font is included in EPSF files, which may in turn be included in
     * PostScript documents (originally CUCPS posters and flyers), we do
     * just this.  Note that this font definition as-is needs to go in the
     * script of an EPSF file, because of the extraneous things it defines.
     * Since we aren't using short names like x, problems aren't very likely,
     * but it is best to avoid them anyway.  */
    xputs("/dotapestart 0 def /dotapeend 0 def /punchchar 0 def\n\n"
	  /* Used in BuildGlyph */);
    fontllx = floor(MIN(norm_left, MIN(lend_left, rend_left)));
    fontlly = floor(font_base);
    fonturx = ceil(MAX(norm_right, MAX(lend_right, rend_right)));
    fontury = ceil(font_top);
    normllx = floor(norm_left);
    normurx = ceil(norm_right);
    l_urx = ceil(lend_right);
    r_urx = ceil(rend_right);
    xprintf("8 dict begin\n"
	    "\n"
	    "/FontType 3 def\n"
	    "/FontMatrix [ .001 0 0 .001 0 0 ] def\n"
	    "/FontBBox [ %d %d %d %d ] def\n"
	    "\n",
	    fontlly, fontlly, fonturx, fontury);
    xputs("/Encoding 256 array def\n"
	  "0 1 255 { Encoding exch dup 3 string cvs cvn put } for\n"
	  "\n");
    xputs("/BuildChar {\n"
	  "  1 index /Encoding get exch get\n"
	  "  1 index /BuildGlyph get exec\n"
	  "} bind def\n"
	  "\n"
	  "/BuildGlyph {\n"
	  "  dup /.notdef eq\n"
	  "    { pop pop }\n"
	  "    { dup /tapestart eq\n"
	  "      { pop /CharProcs get begin dotapestart end }\n"
	  "      { dup /tapeend eq\n"
	  "        { pop /CharProcs get begin dotapeend end }\n"
	  "        % Otherwise we assume it's a number\n"
	  "        { 3 string cvs cvi exch /CharProcs get begin\n"
	  "          punchchar end }\n"
	  "      ifelse }\n"
	  "    ifelse }\n"
	  "  ifelse\n"
	  "} bind def\n"
	  "\n");
    xprintf("/CharProcs 12 dict def\n"
	    "CharProcs begin\n"
	    "  /punchhole { newpath\n"
	    "               3 -1 roll %.6g add\n"
	    "               3 -1 roll %.6g mul %.6g add\n"
	    "               3 -1 roll %% x y r\n"
	    "               0 360 arc closepath fill\n"
	    "  } bind def\n",
	    norm_width / 2,
	    track_spacing,
	    sprocket_height - track_spacing * pstp->holes_below_sprocket);
    xprintf("  /sprocket { 0 %d %.6g punchhole } bind def\n"
	    "  /punchbit { dup %d ge { 1 add } if %.6g exch %.6g punchhole\n"
	    "  } bind def\n",
	    pstp->holes_below_sprocket,
	    sprocket_radius,
	    pstp->holes_below_sprocket,
	    data_hole_offset,
	    data_radius);
    xprintf("  /punchchar { %.6g 0 %d %d %d %d setcachedevice\n"
	    "               %.6g setlinewidth 0 setlinecap [ ] 0 setdash\n"
	    "               newpath\n"
	    "               0 0 moveto %.6g 0 lineto stroke\n"
	    "               0 1000 moveto %.6g 1000 lineto stroke\n"
	    "               sprocket\n"
	    "               1 0 1 %d { 3 copy pop and\n"
	    "                          0 gt { punchbit } { pop } ifelse\n"
	    "                          2 mul } for pop pop\n"
	    "  } bind def\n",
	    norm_width, normllx, fontlly, normurx, fontury,
	    line_thickness, norm_width, norm_width, pstp->character_bits - 1);
    xprintf("  /dotapestart { %.6g 0 0 %d %d %d setcachedevice\n"
	    "                 %.6g setlinewidth 0 setlinecap [ ] 0 setdash\n"
	    "                 newpath\n"
	    "                 %.6g 0 moveto %.6g 0 lineto %.6g %.6g lineto\n"
	    "                 %.6g 1000 lineto %.6g 1000 lineto\n"
	    "                 gsave stroke grestore %% draw edge\n"
	    "                 closepath clip\n"
	    "                 %d { sprocket %.6g 0 translate } repeat\n"
	    "  } bind def\n",
	    lend_width, fontlly, l_urx, fontury, line_thickness,
	    lend_right, ((norm_width / 2.0)
			 + (norm_width * lend_left_pos)
			 + sprocket_height / tan(dtheta)),
	    norm_width / 2.0 + norm_width * lend_left_pos,
	    sprocket_height,
	    ((norm_width / 2.0) + (norm_width * lend_left_pos)
	     + (1000.0 - sprocket_height) / tan(utheta)), lend_right,
	    lend_char_positions, norm_width);
    xprintf("  /dotapeend { %.6g 0 0 %d %d %d setcachedevice\n"
	    "               %.6g setlinewidth 0 setlinecap [ ] 0 setdash\n"
	    "               newpath\n"
	    "               0 0 moveto %.6g 0 lineto %.6g %.6g lineto\n"
	    "               %.6g 1000 lineto 0 1000 lineto\n"
	    "               gsave stroke grestore closepath clip\n"
	    "               %d { sprocket %.6g 0 translate } repeat\n"
	    "  } bind def\n",
	    rend_width, fontlly, r_urx, fontury, line_thickness,
	    ((norm_width / 2.0) + (norm_width * rend_left_pos)
	     + sprocket_height / tan(dtheta)),
	    norm_width / 2.0 + norm_width * rend_left_pos,
	    sprocket_height,
	    ((norm_width / 2.0) + (norm_width * rend_left_pos)
	     + (1000.0 - sprocket_height) / tan(utheta)),
	    rend_char_positions, norm_width);
    xprintf("end\n" /* CharProcs */
	    "\n"
	    "currentdict\n"
	    "end\n"
	    "/PaperTape%d exch definefont pop\n"
	    "\n",
	    pstp->character_bits);
    xprintf("/scaletape { /PaperTape%d findfont exch scalefont setfont\n"
	    "} bind def\n"
	    "\n"
	    "/scaletext { /Courier findfont exch scalefont /Courierfont\n"
	    "  exch def } bind def\n",
	    pstp->character_bits);
  }
  /* The actual message */
  {
    xprintf("%.6g scaletape\n", MM_TO_PT(pstp->tape_width));
    if (show_text_flag)
      xprintf("%.6g scaletext\n", MM_TO_PT(pstp->inter_character_space) / 0.6);
    xputs("0 0 moveto /tapestart glyphshow\n");
    if (show_text_flag) {
      size_t i;
      xputs("gsave Courierfont setfont\n"
	    "0 ( ) stringwidth pop 0.6 div neg rmoveto\n<");
      for (i = 0; i < ps_number_of_chars; i++) {
	xprintf("%02x", (unsigned int)ps_chars[i]);
	if (i % 32 == 31)
	  xputc((unsigned char)'\n', stdout, "standard output");
      }
      xputs("> show grestore\n");
    }
    {
      size_t i;
      xprintf("<");
      for (i = 0; i < ps_number_of_chars; i++) {
	xprintf("%02x", (unsigned int)ps_bits[i]);
	if (i % 32 == 31)
	  xputc((unsigned char)'\n', stdout, "standard output");
      }
      xputs("> show\n");
    }
    xputs("/tapeend glyphshow\n"
	  "end\n"
	  "%%EOF\n");
  }
  xflush(stdout, "standard output");
}
