/* tape-vert.c - format input as paper tape, vertical ASCII art.  */
/* RCS: $Id: tape-vert.c,v 1.4 1999/01/11 00:24:34 jsm28 Exp $ */

/* Copyright (c) 1998, 1999 Joseph Samuel Myers.  All rights reserved.
 *
 * This file is part of nppt, a program to format its input in a
 * representation of paper tape.
 *
 * Nppt is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * Nppt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with nppt; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <ctype.h>
#include <stdio.h>

#include "nppt.h"
#include "util.h"

struct vert_tape_params {
  const char *top; /* Top bit */
  const char *noholes; /* Basic string for each row (no holes) */
  const char *holes; /* Basic string for each row (holes) */
  const char *bottom; /* Bottom bit */
  size_t len; /* Number of characters in each row */
  const size_t *hole_bits; /* Bit for each place */
  size_t text_place; /* Characters from right that the text can be displayed */
};

static const size_t bsd_tape_bits[] = { 7, 6, 5, 4, 3, 2, 1, 0 };
static const struct vert_tape_params bsd_tape_params = {
  "___________\n",
  "|     .   |\n",
  "|ooooo.ooo|\n",
  "___________\n",
  12,
  bsd_tape_bits,
  1
};

static void vert_start_tape(void);
static void vert_print_char(unsigned int, int);
static void vert_end_tape(void);

const struct tape_type vert_tape_types[] = {
  { "BSD", "Traditional BSD/V7 ppt(6) format 8-track tape", 255U, 8U,
    { vert_start_tape, vert_print_char, vert_end_tape },
    &bsd_tape_params
  },
  { NULL, NULL, 0U, 0U,
    { NULL, NULL, NULL },
    NULL
  }
};

static const struct vert_tape_params *vtp;

/* vert_start_tape - start printing the tape, vertical ASCII art */
static void
vert_start_tape(void)
{
  vtp = (const struct vert_tape_params *)tape_params;
  if (override_screen_columns_flag)
    warning("screen columns specification ignored for this tape design");
  xputs(vtp->top);
}

/* vert_print_char - print a character, vertical ASCII art */
static void
vert_print_char(unsigned int c, int character)
{
  size_t p, q;
  q = 0;
  for (p = 0; p < vtp->len; p++) {
    int c1, c2, pc;
    if (show_text_flag && p == (vtp->len - vtp->text_place)) {
      int pc2;
      pc2 = (isprint(character) ? character : (unsigned char)' ');
      xputc(pc2, stdout, "standard output");
    }
    c1 = (unsigned char)vtp->noholes[p];
    c2 = (unsigned char)vtp->holes[p];
    if (c1 == c2) {
      pc = c1;
    } else {
      pc = ((c & (1 << vtp->hole_bits[q++])) ? c2 : c1);
    }
    xputc(pc, stdout, "standard output");
  }
}

/* vert_end_tape - end printing the tape, vertical ASCII art */
static void
vert_end_tape(void)
{
  xputs(vtp->bottom);
  xflush(stdout, "standard output");
}
