/* util.c - utility functions. */
/* RCS: $Id: util.c,v 1.4 1999/01/11 00:24:34 jsm28 Exp $ */

/* Copyright (c) 1997, 1998, 1999 Joseph Samuel Myers.  All rights reserved.
 *
 * This file is part of nppt, a program to format its input in a
 * representation of paper tape.
 *
 * Nppt is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * Nppt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with nppt; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <errno.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "util.h"

const char *program_name; /* Name of the program */

/* get_program_name - get the program name given argv[0] */
void
get_program_name(const char *arg0)
{
  if (arg0 == NULL) {
    program_name = "(no program name available)";
  } else {
    program_name = strrchr(arg0, '/');
    if (program_name != NULL)
      program_name++;
    else
      program_name = arg0;
  }
}

/* die - print an error (with program name and "\n") and exit */
void
die(const char *format, ...)
{
  va_list args;
  /* We do not check for errors here when writing error messages, since
   * there's nothing useful we could do if we detected one.  */
  fprintf(stderr, "%s: ", program_name);
  va_start(args, format);
  vfprintf(stderr, format, args);
  fprintf(stderr, "\n");
  va_end(args);
  exit(EXIT_FAILURE);
}

/* xvfprintf - print something, checking for errors.  */
void
xvfprintf(FILE *stream, const char *filename, const char *format, va_list args)
{
  int ret;
  ret = vfprintf(stream, format, args);
  if (ret < 0)
    die("writing %s: %s", filename, strerror(errno));
}

/* xfprintf - print something, checking for errors.  */
void
xfprintf(FILE *stream, const char *filename, const char *format, ...)
{
  va_list args;
  va_start(args, format);
  xvfprintf(stream, filename, format, args);
  va_end(args);
}

/* xprintf - print something, checking for errors.  */
void
xprintf(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  xvfprintf(stdout, "standard output", format, args);
  va_end(args);
}

/* xfputs - print a string, checking for errors.  */
void
xfputs(const char *string, FILE *stream, const char *filename)
{
  int ret;
  ret = fputs(string, stream);
  if (ret == EOF)
    die("writing %s: %s", filename, strerror(errno));
}

/* xputs - print a string, checking for errors.  Does NOT add a newline.  */
void
xputs(const char *string)
{
  xfputs(string, stdout, "standard output");
}

/* xputc - print a character, checking for errors.  */
void
xputc(int c, FILE *stream, const char *filename)
{
#ifdef HAVE_PUTC_UNLOCKED
  if (putc_unlocked(c, stream) != c)
    die("%s: error writing file: %s", filename, strerror(errno));
#else
  if (putc(c, stream) != c)
    die("%s: error writing file: %s", filename, strerror(errno));
#endif
}

/* xflush - flush a stream, checking for errors.  */
void
xflush(FILE *stream, const char *filename)
{
  fflush(stream);
  if (ferror(stream))
    die("%s: error writing file: %s", filename, strerror(errno));
}

/* warning - print a warning (with program name and "\n") */
void
warning(const char *format, ...)
{
  va_list args;
  xfprintf(stderr, "standard error", "%s: warning: ", program_name);
  va_start(args, format);
  xvfprintf(stderr, "standard error", format, args);
  xfprintf(stderr, "standard error", "\n");
  va_end(args);
}

/* xmalloc - safely allocate memory */
void *
xmalloc(size_t size)
{
  void *tmp;
  tmp = malloc(size);
  if (tmp == NULL)
    die("out of memory in malloc(%lu)", (unsigned long)size);
  return tmp;
}

/* xrealloc - safely reallocate memory */
void *
xrealloc(void *buf, size_t size)
{
  void *tmp;
  tmp = realloc(buf, size);
  if (tmp == NULL)
    die("out of memory in realloc(%lu)", (unsigned long)size);
  return tmp;
}

/* xcalloc - safely allocate zeroed memory */
void *
xcalloc(size_t size)
{
  void *tmp;
  tmp = calloc(size, 1);
  if (tmp == NULL)
    die("out of memory in calloc(%lu)", (unsigned long)size);
  return tmp;
}

/* filesize - size of a file, or die on error */
off_t
filesize(const char *file)
{
  struct stat buf;
  int ret;
  ret = stat(file, &buf);
  if (ret)
    die("%s: stat failed: %s", file, strerror(errno));
  return buf.st_size;
}
