/* winsize.c - get the window size.  */
/* RCS: $Id: winsize.c,v 1.3 1998/12/18 12:43:56 jsm28 Exp $ */

/* Copyright (c) Joseph Samuel Myers 1998.  All rights reserved.
 *
 * This file is part of nppt, a program to format its input in a
 * representation of paper tape.
 *
 * Nppt is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * Nppt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with nppt; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "util.h"
#include "winsize.h"

static int numeric_getenv(const char *var, unsigned long *value);

/* Get a numeric value from an environment variable, and warn if the value
 * is negative, nonsense, etc.; return -1 on error or value not found, 0
 * otherwise.  */
static int
numeric_getenv(const char *var, unsigned long *value)
{
  char *cvalue;
  cvalue = getenv(var);
  if (cvalue == NULL) {
    return -1;
  }
  if (strtol(cvalue, NULL, 0) < 0) {
    warning("a negative value for $%s does not make sense", var);
    return -1;
  } else {
    char *endptr;
    errno = 0;
    *value = strtoul(cvalue, &endptr, 0);
    if (errno != 0) {
      warning("overflow in value of $%s", var);
      /* We still return the value ULONG_MAX as valid */
    }
    if (endptr == cvalue) {
      warning("non-numeric value for $%s", var);
      return -1;
    }
    if (*endptr != '\0') {
      warning("trailing junk in $%s", var);
    }
    return 0;
  }
}

/* Get the window size if possible.  Warn if an error other than ENOTTY
 * or EINVAL occurs.  We use (a) TIOCGWINSZ; (b) environment variables;
 * (c) defaults of 80x24 if a terminal, 0x0 otherwise.  */
void
get_window_size(struct window_size *win)
{
  win->rows = 0;
  win->cols = 0;
#ifdef TIOCGWINSZ
  {
    struct winsize ws;
    int err;
    err = ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws);
    if (err == -1) {
      switch (errno) {
      case ENOTTY:
      case EINVAL:
	/* Probably not a terminal */
	break;
      default:
	warning("ioctl(stdout, TIOCGWINSZ): %s", strerror(errno));
	break;
      }
    } else {
      win->rows = ws.ws_row;
      win->cols = ws.ws_col;
    }
  }
#endif /* TIOCGWINSZ defined */
  if (win->rows == 0) {
    int ret;
    ret = numeric_getenv("ROWS", &win->rows);
    if (ret == -1) {
      win->rows = (isatty(STDOUT_FILENO) ? 24 : 0);
    }
  }
  if (win->cols == 0) {
    int ret;
    ret = numeric_getenv("COLUMNS", &win->cols);
    if (ret == -1) {
      win->cols = (isatty(STDOUT_FILENO) ? 80 : 0);
    }
  }
}
